"""Supporting code for the paper X.

In general arguments of quantities with units are in SI units with no prefixes.

Modules:
    analytical: Functions implementing the analytical model
    fem_parsing: Functions to parse output from the related FEM package
    plots: Plot type classes and associated helper functions
"""

from actinrings._version import __version__
